# flake8: noqa
from .abstract_consumer import ITaskConsumerFactory
from .admin_factory import AdminFactory, IAdminFactory
from .emulation_factory import EmulationFactory, IEmulationFactory
from .publication_factory import IPublicationFactory, PublicationFactory
from .test_execution_factory import ITestExecutionFactory, TestExecutionFactory



import abc
from typing import cast

from overhave import db
from overhave.entities import FeatureTypeModel


class IFeatureTypeStorage(abc.ABC):
    """ Abstract class for feature type storage. """

    @abc.abstractmethod
    def get_default_feature_type(self) -> FeatureTypeModel:
        pass


class FeatureTypeStorage(IFeatureTypeStorage):
    """ Class for feature type storage. """

    def get_default_feature_type(self) -> FeatureTypeModel:
        with db.create_session() as session:
            feature_type: db.FeatureType = session.query(db.FeatureType).order_by(db.FeatureType.id.asc()).first()
            return cast(FeatureTypeModel, FeatureTypeModel.from_orm(feature_type))



from overhave.factory.base_factory import BaseOverhaveFactory
from overhave.factory.context import TApplicationContext


class FactoryWithS3ManagerInit(BaseOverhaveFactory[TApplicationContext]):
    """ Factory with :class:`S3Manager` initialization. """

    def set_context(self, context: TApplicationContext) -> None:
        super().set_context(context)  # type: ignore
        if not self.context.s3_manager_settings.enabled:
            return
        self._s3_manager.initialize()



# flake8: noqa
from .base_factory import IOverhaveFactory
from .components import (
    AdminFactory,
    EmulationFactory,
    IAdminFactory,
    IEmulationFactory,
    IPublicationFactory,
    ITaskConsumerFactory,
    ITestExecutionFactory,
    PublicationFactory,
    TestExecutionFactory,
)
from .consumer_factory import ConsumerFactory
from .context import (
    OverhaveAdminContext,
    OverhaveEmulationContext,
    OverhavePublicationContext,
    OverhaveTestExecutionContext,
    TApplicationContext,
)
from .getters import get_admin_factory, get_emulation_factory, get_publication_factory, get_test_execution_factory



import abc
from typing import Optional, cast

from overhave import db
from overhave.entities import ScenarioModel


class IScenarioStorage(abc.ABC):
    """ Abstract class for feature type storage. """

    @abc.abstractmethod
    def get_scenario(self, scenario_id: int) -> Optional[ScenarioModel]:
        pass


class ScenarioStorage(IScenarioStorage):
    """ Class for feature type storage. """

    def get_scenario(self, scenario_id: int) -> Optional[ScenarioModel]:
        with db.create_session() as session:
            scenario: Optional[db.Scenario] = session.query(db.Scenario).get(scenario_id)
            if scenario is not None:
                return cast(ScenarioModel, ScenarioModel.from_orm(scenario))
            return None



import abc
from functools import cached_property

from overhave.factory.base_factory import BaseOverhaveFactory, IOverhaveFactory
from overhave.factory.components.abstract_consumer import ITaskConsumerFactory
from overhave.factory.context import OverhavePublicationContext
from overhave.publication import IVersionPublisher, StashVersionPublisher
from overhave.transport import PublicationTask, StashHttpClient


class IPublicationFactory(IOverhaveFactory[OverhavePublicationContext], ITaskConsumerFactory[PublicationTask], abc.ABC):
    """ Abstract factory for Overhave publication application. """


class PublicationFactory(BaseOverhaveFactory[OverhavePublicationContext], IPublicationFactory):
    """ Factory for Overhave publication application. """

    context_cls = OverhavePublicationContext

    @cached_property
    def _stash_client(self) -> StashHttpClient:
        return StashHttpClient(settings=self.context.stash_client_settings)

    @cached_property
    def _stash_publisher(self) -> IVersionPublisher:
        return StashVersionPublisher(
            file_settings=self.context.file_settings,
            project_settings=self.context.project_settings,
            feature_storage=self._feature_storage,
            scenario_storage=self._scenario_storage,
            test_run_storage=self._test_run_storage,
            draft_storage=self._draft_storage,
            file_manager=self._file_manager,
            stash_publisher_settings=self.context.stash_publisher_settings,
            client=self._stash_client,
        )

    def process_task(self, task: PublicationTask) -> None:
        return self._stash_publisher.publish_version(task)



import enum
import logging
from typing import Any, Callable, Dict, Optional

import _pytest
import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Argument, Parser
from _pytest.fixtures import FixtureRequest
from _pytest.main import Session
from _pytest.nodes import Item
from _pytest.python import Function
from _pytest.reports import TestReport
from _pytest.runner import CallInfo
from pydantic import ValidationError
from pydantic.dataclasses import dataclass
from pytest_bdd.parser import Feature, Scenario, Step

from overhave.factory import IAdminFactory
from overhave.pytest_plugin.helpers import (
    add_issue_links_to_report,
    add_scenario_title_to_report,
    get_description_manager,
    get_full_step_name,
    get_scenario,
    get_step_context_runner,
    has_issue_links,
    is_pytest_bdd_item,
    set_issue_links,
)
from overhave.pytest_plugin.proxy_manager import get_proxy_manager

logger = logging.getLogger(__name__)


class StepNotFoundError(RuntimeError):
    """ Exception for situation with missing or incorrect step definition. """


class _OptionName(str, enum.Enum):
    ENABLE_INJECTION = "--enable-injection"

    @property
    def as_variable(self) -> str:
        return self.value.lstrip("--").replace("-", "_")


_ENABLE_INJECTION_HELP = "Injection enabling of Overhave specified objects into PyTest session"
_FACTORY_CONTEXT_HELP = (
    "Relative path to lib with Overhave context definition for it's dynamical resolution before injection"
)


@dataclass(frozen=True)
class _Options:
    enable_injection = Argument(
        _OptionName.ENABLE_INJECTION.value,
        action="store_true",
        dest=_OptionName.ENABLE_INJECTION.as_variable,
        default=False,
        help=_ENABLE_INJECTION_HELP,
    )


_PLUGIN_NAME = "overhave-pytest"
_GROUP_HELP = "Overhave PyTest plugin commands"


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup(_PLUGIN_NAME, _GROUP_HELP)
    group.addoption(*_Options.enable_injection.names(), **_Options.enable_injection.attrs())


def pytest_configure(config: Config) -> None:
    """ Patch pytest_bdd objects in current hook. """
    injection_enabled: bool = config.getoption(_OptionName.ENABLE_INJECTION.as_variable)
    tw = _pytest.config.create_terminal_writer(config)
    if injection_enabled:
        logger.debug("Got %s flag.", _OptionName.ENABLE_INJECTION)
        try:
            logger.debug("Try to patch pytest objects...")
            get_proxy_manager().patch_pytest()
            logger.debug("Successfully patched pytest objects.")
            tw.line("Overhave injector successfully initialized.", green=True)
        except ValidationError as e:
            tw.line(f"Could not initialize Overhave injector!\n{str(e)}", red=True)


def pytest_collection_modifyitems(session: Session) -> None:
    pytest_bdd_scenario_items = (item for item in session.items if is_pytest_bdd_item(item))
    for item in pytest_bdd_scenario_items:
        add_scenario_title_to_report(item)

        links_keyword = get_proxy_manager().factory.context.project_settings.links_keyword
        if isinstance(links_keyword, str):
            set_issue_links(scenario=get_scenario(item), keyword=links_keyword)


def pytest_bdd_before_step(
    request: FixtureRequest, feature: Feature, scenario: Scenario, step: Step, step_func: Callable[[Any], None]
) -> None:
    get_step_context_runner.cache_clear()
    runner = get_step_context_runner()
    runner.set_title(get_full_step_name(step))
    runner.start()


def pytest_bdd_after_step(
    request: FixtureRequest,
    feature: Feature,
    scenario: Scenario,
    step: Step,
    step_func: Callable[[Any], None],
    step_func_args: Dict[str, Any],
) -> None:
    runner = get_step_context_runner()
    runner.stop(None)


def pytest_bdd_step_error(
    request: FixtureRequest,
    feature: Feature,
    scenario: Scenario,
    step: Step,
    step_func: Callable[[Any], None],
    step_func_args: Dict[str, Any],
    exception: BaseException,
) -> None:
    runner = get_step_context_runner()
    runner.stop(exception)


def pytest_bdd_apply_tag(tag: str, function: Function) -> Optional[bool]:
    if tag != "skip":
        return None
    marker = pytest.mark.skip(reason="Scenario manually marked as skipped")
    marker(function)
    return True


def pytest_bdd_step_func_lookup_error(
    request: FixtureRequest, feature: Feature, scenario: Scenario, step: Step, exception: BaseException
) -> None:
    raise StepNotFoundError(f"Could not found specified step '{get_full_step_name(step)}'") from exception


def pytest_collection_finish(session: Session) -> None:
    """ Supplying of injector configs for steps collection. """
    proxy_manager = get_proxy_manager()
    if (
        session.config.getoption(_OptionName.ENABLE_INJECTION.as_variable)
        and proxy_manager.has_factory
        and isinstance(proxy_manager.factory, IAdminFactory)
    ):
        tw = _pytest.config.create_terminal_writer(session.config)
        if not proxy_manager.pytest_patched:
            tw.line("Could not supplement Overhave injector - pytest session has not been patched!", yellow=True)
            return
        try:
            proxy_manager.supply_injector_for_collection()
            tw.line("Overhave injector successfully supplemented.", green=True)
        except ValidationError as e:
            tw.line(f"Could not supplement Overhave injector!\n{str(e)}", red=True)
        proxy_manager.injector.adapt(session)


def pytest_runtest_setup(item: Item) -> None:
    """ Hook for purgation of get_description_manager. """
    get_description_manager.cache_clear()


def pytest_runtest_makereport(item: Item, call: CallInfo[None]) -> Optional[TestReport]:
    """ Hook for description and issue links attachment to Allure report. """
    get_description_manager().apply_description()
    proxy_manager = get_proxy_manager()
    if all((proxy_manager.factory.context.project_settings.browse_url is not None, has_issue_links(item))):
        add_issue_links_to_report(
            project_settings=proxy_manager.factory.context.project_settings, scenario=get_scenario(item)
        )
    return None  # noqa: R501



import logging
import re
from typing import List, Optional

import ldap
from ldap.ldapobject import LDAPObject

from overhave.entities.authorization.settings import OverhaveLdapClientSettings

logger = logging.getLogger(__name__)


class LDAPAuthenticator:
    """ Class-client for LDAP authentication. """

    def __init__(self, settings: OverhaveLdapClientSettings) -> None:
        self._settings = settings
        self._ldap_connection: Optional[LDAPObject] = None

    def _connect(self, login: str, password: str) -> None:
        ldap_connection = ldap.initialize(self._settings.ldap_url)
        ldap_connection.set_option(ldap.OPT_REFERRALS, 0)
        ldap_connection.set_option(ldap.OPT_NETWORK_TIMEOUT, self._settings.ldap_timeout.seconds)
        ldap_connection.simple_bind_s(f"{self._settings.ldap_domain}{login}", password)

        self._ldap_connection = ldap_connection

    def _ask_ad_usergroups(self, login: str) -> List[str]:
        result = self._ldap_connection.search_st(  # type: ignore
            base=self._settings.ldap_dn,
            scope=ldap.SCOPE_SUBTREE,
            filterstr=f"(sAMAccountName={login})",
            attrlist=["memberOf"],
            timeout=self._settings.ldap_timeout.seconds,
        )
        member_of = result[0][1]["memberOf"]
        member_of = [m.decode("utf8") for m in member_of]

        p = re.compile("CN=(.*?),", re.IGNORECASE)
        return [
            p.match(x).group(1)  # type: ignore
            for x in list(filter(lambda x: "OU=Security Groups" in x or "OU=Mail Groups" in x, member_of))
        ]

    def get_user_groups(self, login: str, password: str) -> Optional[List[str]]:
        try:
            self._connect(login, password)
        except ldap.INVALID_CREDENTIALS:
            logger.info("Failed LDAP authorization for user: %s", login)
            return None

        return self._ask_ad_usergroups(login)



from typing import List

import allure
from markupsafe import Markup

from overhave.entities import OverhaveDescriptionManagerSettings


class DescriptionManager:
    """ Class for test-suit custom description management and setting to Allure report. """

    def __init__(self, settings: OverhaveDescriptionManagerSettings):
        self._settings = settings
        self._description: List[str] = []

    def apply_description(self) -> None:
        if self._description:
            joined_description = self._settings.blocks_delimiter.join(self._description)
            allure.dynamic.description_html(Markup(joined_description))

    def add_description(self, value: str) -> None:
        self._description.append(value)

    def add_description_above(self, value: str) -> None:
        self._description.insert(0, value)



import abc
from functools import cached_property
from typing import Any, Dict

from overhave.entities import ReportManager
from overhave.entities.authorization import AuthorizationStrategy, IAdminAuthorizationManager, LDAPAuthenticator
from overhave.entities.authorization.mapping import AUTH_STRATEGY_TO_MANAGER_MAPPING
from overhave.factory.base_factory import IOverhaveFactory
from overhave.factory.components.s3_init_factory import FactoryWithS3ManagerInit
from overhave.factory.context import OverhaveAdminContext
from overhave.storage import IFeatureTypeStorage
from overhave.transport import EmulationTask, PublicationTask, RedisProducer, RedisStream, TestRunTask


class IAdminFactory(IOverhaveFactory[OverhaveAdminContext]):
    """ Factory interface for Overhave admin application. """

    @property
    @abc.abstractmethod
    def auth_manager(self) -> IAdminAuthorizationManager:
        pass

    @property
    @abc.abstractmethod
    def feature_type_storage(self) -> IFeatureTypeStorage:
        pass

    @property
    @abc.abstractmethod
    def redis_producer(self) -> RedisProducer:
        pass

    @property
    @abc.abstractmethod
    def report_manager(self) -> ReportManager:
        pass


class AdminFactory(FactoryWithS3ManagerInit[OverhaveAdminContext], IAdminFactory):
    """ Factory for Overhave admin application. """

    context_cls = OverhaveAdminContext

    @cached_property
    def _auth_manager(self) -> IAdminAuthorizationManager:
        settings = self.context.auth_settings
        kwargs: Dict[str, Any] = dict(settings=settings)
        if settings.auth_strategy is AuthorizationStrategy.LDAP:
            kwargs["ldap_authenticator"] = LDAPAuthenticator(settings=self.context.ldap_client_settings)
        return AUTH_STRATEGY_TO_MANAGER_MAPPING[settings.auth_strategy](**kwargs)  # type: ignore

    @property
    def auth_manager(self) -> IAdminAuthorizationManager:
        return self._auth_manager

    @property
    def feature_type_storage(self) -> IFeatureTypeStorage:
        return self._feature_type_storage

    @cached_property
    def _redis_producer(self) -> RedisProducer:
        return RedisProducer(
            settings=self.context.redis_settings,
            mapping={
                TestRunTask: RedisStream.TEST,
                PublicationTask: RedisStream.PUBLICATION,
                EmulationTask: RedisStream.EMULATION,
            },
        )

    @property
    def redis_producer(self) -> RedisProducer:
        return self._redis_producer



import enum
from typing import TYPE_CHECKING, Optional

import sqlalchemy as sa
import sqlalchemy_utils as su

from overhave.db.base import BaseTable, PrimaryKeyMixin
from overhave.db.types import LONG_STR_TYPE, SHORT_STR_TYPE


class Role(str, enum.Enum):
    """ Enum that declares user access roles. """

    user = "user"
    admin = "admin"


class BaseUser:
    """ Base user class. """

    login: str = sa.Column(SHORT_STR_TYPE, nullable=False, unique=True)
    password: Optional[str] = sa.Column(LONG_STR_TYPE, nullable=True)
    role: str = sa.Column(sa.Enum(Role), nullable=False, default=Role.user)

    if TYPE_CHECKING:
        id: int

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def is_active(self) -> bool:
        return True

    @property
    def is_anonymous(self) -> bool:
        return False

    def get_id(self) -> int:
        return self.id

    # Required for administrative interface
    def __unicode__(self) -> str:
        return self.login

    def __str__(self) -> str:
        return self.login


@su.generic_repr("login")
class UserRole(BaseTable, PrimaryKeyMixin, BaseUser):
    """ User access table. """

    def __init__(self, login: str, password: Optional[str], role: Role) -> None:
        self.login = login
        self.password = password
        self.role = role


@su.generic_repr("group")
class GroupRole(BaseTable, PrimaryKeyMixin):
    """ Group access table. """

    group = sa.Column(SHORT_STR_TYPE, nullable=False, unique=True)

    def __init__(self, group: str) -> None:
        self.group = group



from typing import Optional

from overhave.entities import OverhaveFileSettings, OverhaveLanguageSettings, OverhaveScenarioCompilerSettings
from overhave.factory.context.base_context import BaseFactoryContext
from overhave.publication import OverhaveStashPublisherSettings
from overhave.test_execution import OverhaveProjectSettings
from overhave.transport import OverhaveStashClientSettings


class OverhavePublicationContext(BaseFactoryContext):
    """ Overhave publication context, based on application BaseSettings.

    This context defines how Overhave publication will work.
    """

    def __init__(
        self,
        compilation_settings: Optional[OverhaveScenarioCompilerSettings] = None,
        file_settings: Optional[OverhaveFileSettings] = None,
        language_settings: Optional[OverhaveLanguageSettings] = None,
        project_settings: Optional[OverhaveProjectSettings] = None,
        stash_client_settings: Optional[OverhaveStashClientSettings] = None,
        stash_publisher_settings: Optional[OverhaveStashPublisherSettings] = None,
    ) -> None:
        super().__init__(
            compilation_settings=compilation_settings or OverhaveScenarioCompilerSettings(),
            file_settings=file_settings or OverhaveFileSettings(),
            language_settings=language_settings or OverhaveLanguageSettings(),
            project_settings=project_settings or OverhaveProjectSettings(),
        )
        self.stash_client_settings = stash_client_settings or OverhaveStashClientSettings()
        self.stash_publisher_settings = stash_publisher_settings or OverhaveStashPublisherSettings()



import click
import sqlalchemy_utils as sau
from alembic.config import Config
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import OperationalError

from overhave import db as database
from overhave.base_settings import DataBaseSettings


def _create_all(config: Config) -> None:
    click.echo("creating")
    config.attributes["metadata"].create_all()
    click.echo("complete!")


@click.command(short_help="Create all metadata tables")
@click.pass_obj
def create_all(config: Config) -> None:
    """ Create all metadata tables. """
    _create_all(config)


def _drop_all(config: Config) -> None:
    click.echo("dropping")
    meta = config.attributes["metadata"]
    engine = config.attributes["engine"]
    for table in meta.tables:
        engine.execute(f'DROP TABLE IF EXISTS "{table}" CASCADE')
    engine.execute("DROP TABLE IF EXISTS alembic_version")
    engine.execute("DROP SCHEMA IF EXISTS huey")
    meta.drop_all()
    click.echo("complete!")


@click.command(short_help="Drop all metadata tables, attributes, schema")
@click.pass_obj
def drop_all(config: Config) -> None:
    """ Drop all metadata tables, attributes, schema. """
    click.confirm("it really need?", abort=True)
    _drop_all(config)


def _ensure_database_exists(db_url: URL) -> None:
    try:
        if not sau.database_exists(db_url):
            sau.create_database(db_url)
    except OperationalError as e:
        click.echo(e)
        click.echo("Catched error when trying to check database existence!")


def set_config_to_context(context: click.Context, settings: DataBaseSettings) -> None:
    """ Set Alembic config to Click context for easy operations and migrations ability. """
    _ensure_database_exists(settings.db_url)
    settings.setup_db()
    config = Config()
    config.attributes["engine"] = settings.create_engine()
    config.attributes["metadata"] = database.metadata
    context.obj = config



import abc
from typing import cast

from overhave import db
from overhave.entities import FeatureModel


class IFeatureStorage(abc.ABC):
    """ Abstract class for feature storage. """

    @abc.abstractmethod
    def get_feature(self, feature_id: int) -> FeatureModel:
        pass


class FeatureStorage(IFeatureStorage):
    """ Class for feature storage. """

    def get_feature(self, feature_id: int) -> FeatureModel:
        with db.create_session() as session:
            scenario: db.Feature = session.query(db.Feature).filter(db.Feature.id == feature_id).one()
            return cast(FeatureModel, FeatureModel.from_orm(scenario))



import abc
from pathlib import Path
from typing import Dict, List

from overhave.entities.feature.types import FeatureTypeName


class IFeatureExtractor(abc.ABC):
    """ Abstract class for specified project's feature types resolution. """

    @property
    @abc.abstractmethod
    def feature_types(self) -> List[FeatureTypeName]:
        pass

    @property
    @abc.abstractmethod
    def feature_type_to_dir_mapping(self) -> Dict[FeatureTypeName, Path]:
        pass



# flake8: noqa
from typing import TypeVar

from .admin_context import OverhaveAdminContext
from .emulation_context import OverhaveEmulationContext
from .publication_context import OverhavePublicationContext
from .test_execution_context import OverhaveTestExecutionContext

TApplicationContext = TypeVar(
    "TApplicationContext",
    OverhaveAdminContext,
    OverhaveTestExecutionContext,
    OverhavePublicationContext,
    OverhaveEmulationContext,
)



from functools import cached_property, partial
from typing import Callable, Dict, Type

from overhave.factory.getters import get_emulation_factory, get_publication_factory, get_test_execution_factory
from overhave.pytest_plugin import get_proxy_manager
from overhave.transport import (
    AnyRedisTask,
    EmulationTask,
    PublicationTask,
    RedisConsumer,
    RedisConsumerRunner,
    RedisStream,
    TestRunTask,
)


class ConsumerFactory:
    """ Factory for :class:`RedisConsumer`, :class:`RedisConsumerRunner` and tasks mapping. """

    def __init__(self, stream: RedisStream):
        self._stream = stream

    @cached_property
    def _consumer(self) -> RedisConsumer:
        from overhave.entities import OverhaveRedisSettings

        return RedisConsumer(settings=OverhaveRedisSettings(), stream_name=self._stream)

    @cached_property
    def runner(self) -> RedisConsumerRunner:
        return RedisConsumerRunner(consumer=self._consumer, mapping=self._mapping)

    @cached_property
    def _mapping(self) -> Dict[Type[AnyRedisTask], Callable[[AnyRedisTask], None]]:
        return {
            TestRunTask: self._process_test_execution_task,  # type: ignore
            PublicationTask: get_publication_factory().process_task,  # type: ignore
            EmulationTask: get_emulation_factory().process_task,  # type: ignore
        }

    @cached_property
    def _process_test_execution_task(self) -> Callable[[TestRunTask], None]:
        factory = get_test_execution_factory()
        proxy_manager = get_proxy_manager()
        proxy_manager.set_factory(factory)
        return partial(factory.process_task,)



import abc

from overhave.entities import OverhaveFileSettings, PublisherContext
from overhave.publication.abstract_publisher import IVersionPublisher
from overhave.scenario import FileManager, generate_task_info
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.test_execution import OverhaveProjectSettings


class BaseVersionPublisherException(Exception):
    """ Base exception for :class:`BaseVersionPublisher`. """


class DraftNotExistsError(BaseVersionPublisherException):
    """ Exception for situation with not existing Draft. """


class TestRunNotExistsError(BaseVersionPublisherException):
    """ Exception for situation with not existing TestRun. """


class ScenarioNotExistsError(BaseVersionPublisherException):
    """ Exception for situation with not existing Scenario. """


class BaseVersionPublisher(IVersionPublisher, abc.ABC):
    """ Class for feature version's pull requests management relative to Atlassian Stash API. """

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
    ) -> None:
        self._file_settings = file_settings
        self._project_settings = project_settings
        self._feature_storage = feature_storage
        self._scenario_storage = scenario_storage
        self._test_run_storage = test_run_storage
        self._draft_storage = draft_storage
        self._file_manager = file_manager

    def _compile_context(self, draft_id: int) -> PublisherContext:
        draft = self._draft_storage.get_draft(draft_id)
        if not draft:
            raise DraftNotExistsError(f"Draft with id={draft_id} does not exist!")
        test_run = self._test_run_storage.get_test_run(draft.test_run_id)
        if not test_run:
            raise TestRunNotExistsError(f"TestRun with id={draft.test_run_id} does not exist!")
        feature = self._feature_storage.get_feature(draft.feature_id)
        scenario = self._scenario_storage.get_scenario(test_run.scenario_id)
        if not scenario:
            raise ScenarioNotExistsError(f"Scenario with id={test_run.scenario_id} does not exist!")
        return PublisherContext(
            feature=feature,
            scenario=scenario,
            test_run=test_run,
            draft=draft,
            target_branch=f"bdd-feature-{feature.id}",
        )

    def _compile_publication_description(self, context: PublisherContext) -> str:
        return "\n".join(
            (
                f"Feature ID: {context.feature.id}. Type: '{context.feature.feature_type.name}'.",
                f"Created by: @{context.feature.author} at {context.feature.created_at.strftime('%d-%m-%Y %H:%M:%S')}.",
                f"Last edited by: @{context.feature.last_edited_by}.",
                f"PR from Test Run ID: {context.test_run.id}. Executed by: @{context.test_run.executed_by}",
                f"Published by: @{context.draft.published_by}.",
                generate_task_info(tasks=context.feature.task, header=self._project_settings.links_keyword),
            )
        )



import abc
from functools import cached_property

from overhave.entities import Emulator
from overhave.factory.base_factory import BaseOverhaveFactory, IOverhaveFactory
from overhave.factory.components.abstract_consumer import ITaskConsumerFactory
from overhave.factory.context import OverhaveEmulationContext
from overhave.transport import EmulationTask


class IEmulationFactory(IOverhaveFactory[OverhaveEmulationContext], ITaskConsumerFactory[EmulationTask], abc.ABC):
    """ Abstract factory for Overhave emulation application. """


class EmulationFactory(BaseOverhaveFactory[OverhaveEmulationContext], IEmulationFactory):
    """ Factory for Overhave emulation application. """

    context_cls = OverhaveEmulationContext

    @cached_property
    def _emulator(self) -> Emulator:
        return Emulator(settings=self.context.emulation_settings, storage=self._emulation_storage)

    def process_task(self, task: EmulationTask) -> None:
        return self._emulator.start_emulation(task)



import click

from overhave import overhave_app
from overhave.base_settings import DataBaseSettings, LoggingSettings
from overhave.cli.group import overhave
from overhave.factory import get_admin_factory


def _run_admin(port: int, debug: bool) -> None:
    DataBaseSettings().setup_db()
    LoggingSettings().setup_logging()
    overhave_app(get_admin_factory()).run(host="0.0.0.0", port=port, debug=debug)


@overhave.command(short_help="Run Overhave web-service")
@click.option("--port", default=8076)
@click.option("--debug", default=False)
def admin(port: int, debug: bool) -> None:
    _run_admin(port, debug)



from __future__ import annotations

import datetime
import logging
import re
from typing import List, Tuple, Type, Union

import sqlalchemy as sa
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import as_declarative, declared_attr
from sqlalchemy.orm import Mapper
from sqlalchemy.orm import Session as SessionClass
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.query import Query

from overhave.db.types import INT_TYPE

logger = logging.getLogger(__name__)

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)


@as_declarative(metadata=metadata)
class BaseTable:
    """ Base table class with __tablename__. """

    @declared_attr
    def __tablename__(cls) -> str:
        return _classname_to_tablename(cls.__name__)  # type: ignore


class PrimaryKeyWithoutDateMixin:
    """ Table mixin with declared attribute `id`. """

    @declared_attr
    def id(cls) -> sa.Column[int]:
        return sa.Column(f"{cls.__tablename__}_id", INT_TYPE, primary_key=True)  # type: ignore


class PrimaryKeyMixin(PrimaryKeyWithoutDateMixin):
    """ Table mixin with `id` and `created_at` declared attributes. """

    @declared_attr
    def created_at(cls) -> sa.Column[datetime.datetime]:
        return sa.Column(sa.DateTime(timezone=True), nullable=True, server_default=sa.func.now())


def _get_query_cls(mapper: Union[Tuple[Type[BaseTable], ...], Mapper], session: SessionClass) -> Query:
    if mapper:
        m = mapper
        if isinstance(m, tuple):
            m = mapper[0]
        if isinstance(m, Mapper):
            m = m.entity

        try:
            return m.__query_cls__(mapper, session)
        except AttributeError:
            pass

    return Query(mapper, session)


def _classname_to_tablename(name: str) -> str:
    result: List[str] = []

    last_index = 0
    for match in re.finditer(r"(?P<abbreviation>[A-Z]+(?![a-z\d]))|(?P<word>[A-Z][a-z]*)|(?P<digit>\d+)", name):
        if match.start() != last_index:
            raise ValueError(f'Could not translate class name "{name}" to table name')

        last_index = match.end()
        result.append(match.group().lower())

    return "_".join(result)


Session = sessionmaker(query_cls=_get_query_cls)

current_session = scoped_session(Session)



import abc
import logging
from uuid import UUID, uuid4

from overhave import db
from overhave.entities.authorization.manager.base import BaseAdminAuthorizationManager
from overhave.entities.authorization.settings import OverhaveAuthorizationSettings

logger = logging.getLogger(__name__)
_ADMIN_USERNAME = "admin"


class AdminSecretMixin(BaseAdminAuthorizationManager, abc.ABC):
    """
    Special class for default admin user preparation.

    Usage: application will create 'admin' user with UUID randomly generated secret key.
    This user will be placed into database, moreover key will be logged with INFO logging level.
    Note: real service administrator should bring this key and change password or delete this user!
    """

    def __init__(self, settings: OverhaveAuthorizationSettings):
        super().__init__(settings)
        self._make_secret()

    @staticmethod
    def _get_secret() -> UUID:
        return uuid4()

    @classmethod
    def _make_secret(cls) -> None:
        with db.create_session(expire_on_commit=False) as session:
            db_user = session.query(db.UserRole).filter(db.UserRole.login == _ADMIN_USERNAME).one_or_none()
            if db_user is not None:
                logger.info("Admin user already exists")
                return
            secret = cls._get_secret()
            logger.info("Generated admin secret: %s", secret)
            cls._create_user(session=session, username=_ADMIN_USERNAME, password=str(secret), role=db.Role.admin)



from pathlib import Path
from shutil import make_archive, unpack_archive

from overhave.entities.settings import OverhaveFileSettings


class BaseArchiveManagerException(Exception):
    """ Base exception for :class:`ArchiveManager`. """


class IncorrectExtensionError(BaseArchiveManagerException):
    """ Exception for situation with specifing incorrect file extension for unpacking. """


class ArchiveManager:
    """ Class for files archiving management. """

    def __init__(self, file_settings: OverhaveFileSettings):
        self._file_settings = file_settings

    def archive_path(self, path: Path, extension: str) -> Path:
        zipped = self._file_settings.tmp_reports_dir / path.name
        return Path(make_archive(zipped.as_posix(), extension, path))

    def unpack_path(self, path: Path, extension: str) -> Path:
        if not path.name.endswith(extension):
            raise IncorrectExtensionError(
                "Incorrect file extension '%s' for file '%s' unpacking!", extension, path.name
            )
        extract_path = self._file_settings.tmp_reports_dir / path.name.replace(f".{extension}", "")
        unpack_archive(path.as_posix(), extract_path.as_posix(), extension)
        return extract_path



import logging
from typing import Optional, cast

from wtforms import PasswordField, StringField

from overhave import db
from overhave.entities.authorization.manager.secret_mixin import AdminSecretMixin

logger = logging.getLogger(__name__)


class SimpleAdminAuthorizationManager(AdminSecretMixin):
    """ Class for user registration via database.

    Manager does not provide real authorization. Every user could use preferred name.
    This name will be used for user authority. Every user is unique. Passwords not required.
    """

    def authorize_user(self, username_field: StringField, password_field: PasswordField) -> Optional[db.BaseUser]:
        with db.create_session(expire_on_commit=False) as s:
            db_user = s.query(db.UserRole).filter(db.UserRole.login == username_field.data).one_or_none()
            if db_user is None:
                db_user = self._create_user(session=s, username=username_field.data, password=password_field.data)
            if db_user.password == password_field.data:
                return cast(db.BaseUser, db_user)
            return None



from flask_login import current_user
from wtforms import ValidationError

from overhave.admin.views.base import ModelViewProtected
from overhave.db import Role


class DraftView(ModelViewProtected):
    """ View for :class:`Draft` table. """

    details_template = "draft_detail.html"
    can_delete = True
    column_list = ("id", "feature_id", "test_run_id", "pr_url", "published_by", "created_at")
    column_exclude_list = ("feature", "text")  # type: ignore
    column_details_list = ("id", "feature_id", "test_run_id", "pr_url", "published_by", "created_at")

    column_sortable_list = ("id", "feature_id")
    column_searchable_list = (
        "test_run_id",
        "published_by",
        "pr_url",
    )
    column_filters = ("feature_id", "published_by", "created_at")

    def on_model_delete(self, model) -> None:  # type: ignore
        if not current_user.role == Role.admin:
            raise ValidationError("Only administrator could delete published version of scenario!")



# flake8: noqa
from .auth import BearerAuth
from .client import BaseHttpClient, BaseHttpClientException, HttpClientValidationError
from .objects import HttpMethod
from .settings import BaseHttpClientSettings



import abc
from datetime import datetime
from typing import List, Optional, cast

from overhave import db
from overhave.entities import DraftModel


class IDraftStorage(abc.ABC):
    """ Abstract class for scenario versions storage. """

    @abc.abstractmethod
    def get_draft(self, draft_id: int) -> Optional[DraftModel]:
        pass

    @abc.abstractmethod
    def save_draft(self, test_run_id: int, published_by: str) -> int:
        pass

    @abc.abstractmethod
    def save_response(self, draft_id: int, pr_url: str, published_at: datetime, opened: bool) -> None:
        pass

    @abc.abstractmethod
    def get_previous_feature_draft(self, feature_id: int) -> DraftModel:
        pass


class BaseDraftStorageException(Exception):
    """ Base exception for :class:`DraftStorage`. """


class UniqueDraftCreationError(BaseDraftStorageException):
    """ Exception for draft creation error with `as_unique`. """


class NullableDraftsError(BaseDraftStorageException):
    """ Exception for situation with not existing drafts. """


class DraftStorage(IDraftStorage):
    """ Class for scenario versions storage. """

    def get_draft(self, draft_id: int) -> Optional[DraftModel]:
        with db.create_session() as session:
            draft: Optional[db.Draft] = session.query(db.Draft).get(draft_id)
            if draft is not None:
                return cast(DraftModel, DraftModel.from_orm(draft))
            return None

    def save_draft(self, test_run_id: int, published_by: str) -> int:
        with db.create_session() as session:
            try:
                draft = session.query(db.Draft).as_unique(test_run_id=test_run_id, published_by=published_by)
            except RuntimeError as e:
                raise UniqueDraftCreationError("Could not get unique draft!") from e
            session.add(draft)
            session.flush()
            return cast(int, draft.id)

    def save_response(self, draft_id: int, pr_url: str, published_at: datetime, opened: bool) -> None:
        with db.create_session() as session:
            draft: db.Draft = session.query(db.Draft).get(draft_id)
            draft.pr_url = pr_url
            draft.published_at = published_at
            feature: db.Feature = session.query(db.Feature).get(draft.feature_id)
            feature.released = opened

    def get_previous_feature_draft(self, feature_id: int) -> DraftModel:
        with db.create_session() as session:
            selection_num = 2
            drafts: List[db.Draft] = session.query(db.Draft).filter(db.Draft.feature_id == feature_id).order_by(
                db.Draft.id.asc()
            ).limit(selection_num).all()
            if not drafts or len(drafts) != selection_num:
                raise NullableDraftsError(f"Haven't got Drafts amount={selection_num} for feature_id={feature_id}!")
            return cast(DraftModel, DraftModel.from_orm(drafts[0]))



import sqlalchemy as sa

INT_TYPE = sa.Integer()
LONG_STR_TYPE = sa.String(90)
SHORT_STR_TYPE = sa.String(30)
TEXT_TYPE = sa.Text()
DATETIME_TYPE = sa.DateTime(timezone=True)
ARRAY_TYPE = sa.ARRAY(sa.String(12))



import re
from typing import Any

import sqlalchemy as sa
from sqlalchemy.types import TypeDecorator


class StringWithoutSpecialSymbols(TypeDecorator):
    """ String without any special characters. """

    impl = sa.String

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.__class__.__name__ = "String"
        super().__init__(*args, **kwargs)

    def process_bind_param(self, value: Any, dialect: Any) -> str:
        if isinstance(value, str) and re.match(r"^[a-z0-9A-Zа-яА-ЯёЁ_]+$", value):
            return value
        raise ValueError("Object shouldn`t contain any special symbols or spaces")

    def process_result_value(self, value: Any, dialect: Any) -> Any:
        return value

    def copy(self, **kw: Any) -> Any:
        return StringWithoutSpecialSymbols(self.impl.length)



import abc
from functools import cached_property
from typing import Generic, Optional, Type, cast

from overhave.entities import ArchiveManager, FeatureExtractor, IFeatureExtractor, ReportManager
from overhave.factory.context import TApplicationContext
from overhave.scenario import FileManager, ScenarioCompiler
from overhave.storage import (
    DraftStorage,
    EmulationStorage,
    FeatureStorage,
    FeatureTypeStorage,
    IDraftStorage,
    IEmulationStorage,
    IFeatureStorage,
    IFeatureTypeStorage,
    IScenarioStorage,
    ITestRunStorage,
    ScenarioStorage,
    TestRunStorage,
)
from overhave.test_execution import PytestRunner, StepCollector
from overhave.transport import S3Manager


class IOverhaveFactory(Generic[TApplicationContext], abc.ABC):
    """ Factory interface for application entities resolution and usage. """

    @abc.abstractmethod
    def set_context(self, context: TApplicationContext) -> None:
        pass

    @property
    @abc.abstractmethod
    def context(self) -> TApplicationContext:
        pass

    @property
    @abc.abstractmethod
    def draft_storage(self) -> IDraftStorage:
        pass

    @property
    @abc.abstractmethod
    def emulation_storage(self) -> IEmulationStorage:
        pass

    @property
    @abc.abstractmethod
    def feature_extractor(self) -> IFeatureExtractor:
        pass

    @property
    @abc.abstractmethod
    def scenario_storage(self) -> IScenarioStorage:
        pass

    @property
    @abc.abstractmethod
    def step_collector(self) -> StepCollector:
        pass

    @property
    @abc.abstractmethod
    def test_run_storage(self) -> ITestRunStorage:
        pass

    @property
    @abc.abstractmethod
    def test_runner(self) -> PytestRunner:
        pass


class BaseOverhaveFactory(IOverhaveFactory[TApplicationContext]):
    """ Base factory for application entities resolution and usage. """

    context_cls: Type[TApplicationContext]

    def __init__(self) -> None:
        self._context: Optional[TApplicationContext] = None

    def set_context(self, context: TApplicationContext) -> None:
        self._context = context

    @property
    def context(self) -> TApplicationContext:
        if self._context is None:
            self.set_context(self.context_cls())
        return cast(TApplicationContext, self._context)

    @cached_property
    def _archive_manager(self) -> ArchiveManager:
        return ArchiveManager(file_settings=self.context.file_settings)

    @cached_property
    def _draft_storage(self) -> DraftStorage:
        return DraftStorage()

    @property
    def draft_storage(self) -> IDraftStorage:
        return self._draft_storage

    @cached_property
    def _emulation_storage(self) -> EmulationStorage:
        return EmulationStorage(settings=self.context.emulation_settings)

    @property
    def emulation_storage(self) -> IEmulationStorage:
        return self._emulation_storage

    @cached_property
    def _feature_extractor(self) -> FeatureExtractor:
        return FeatureExtractor(file_settings=self.context.file_settings)

    @property
    def feature_extractor(self) -> IFeatureExtractor:
        return self._feature_extractor

    @cached_property
    def _feature_storage(self) -> IFeatureStorage:
        return FeatureStorage()

    @cached_property
    def _feature_type_storage(self) -> IFeatureTypeStorage:
        return FeatureTypeStorage()

    @cached_property
    def _file_manager(self) -> FileManager:
        for path in (
            self.context.file_settings.tmp_features_dir,
            self.context.file_settings.tmp_fixtures_dir,
            self.context.file_settings.tmp_reports_dir,
        ):
            path.mkdir(parents=True, exist_ok=True)
        return FileManager(
            project_settings=self.context.project_settings,
            file_settings=self.context.file_settings,
            language_settings=self.context.language_settings,
            feature_extractor=self._feature_extractor,
            scenario_compiler=self._scenario_compiler,
        )

    @cached_property
    def _report_manager(self) -> ReportManager:
        return ReportManager(
            settings=self.context.report_manager_settings,
            file_settings=self.context.file_settings,
            test_run_storage=self._test_run_storage,
            archive_manager=self._archive_manager,
            s3_manager=self._s3_manager,
        )

    @property
    def report_manager(self) -> ReportManager:
        return self._report_manager

    @cached_property
    def _s3_manager(self) -> S3Manager:
        return S3Manager(self.context.s3_manager_settings)

    @cached_property
    def _scenario_compiler(self) -> ScenarioCompiler:
        return ScenarioCompiler(
            compilation_settings=self.context.compilation_settings,
            language_settings=self.context.language_settings,
            task_links_keyword=self.context.project_settings.links_keyword,
        )

    @cached_property
    def _scenario_storage(self) -> IScenarioStorage:
        return ScenarioStorage()

    @property
    def scenario_storage(self) -> IScenarioStorage:
        return self._scenario_storage

    @cached_property
    def _step_collector(self) -> StepCollector:
        return StepCollector(step_prefixes=self.context.language_settings.step_prefixes)

    @property
    def step_collector(self) -> StepCollector:
        return self._step_collector

    @cached_property
    def _test_run_storage(self) -> TestRunStorage:
        return TestRunStorage()

    @property
    def test_run_storage(self) -> ITestRunStorage:
        return self._test_run_storage

    @cached_property
    def _test_runner(self) -> PytestRunner:
        return PytestRunner(settings=self.context.test_settings)

    @property
    def test_runner(self) -> PytestRunner:
        return self._test_runner



# flake8: noqa
from .base_client import BaseHttpClient, BaseHttpClientSettings
from .stash_client import (
    OverhaveStashClientSettings,
    StashBranch,
    StashErrorResponse,
    StashHttpClient,
    StashHttpClientConflictError,
    StashPrCreationResponse,
    StashProject,
    StashPrRequest,
    StashRepository,
    StashReviewer,
    StashReviewerInfo,
)



import logging
from operator import attrgetter
from typing import Any, Dict, List, Optional, Tuple, cast

from _pytest.fixtures import FixtureDef
from _pytest.main import Session
from pytest_bdd import types as default_types

from overhave.entities import FeatureTypeName, StepPrefixesModel

_PYTESTBDD_FIXTURE_MARK = "pytestbdd_"
_PYTESTBDD_FIXTURE_TRACE_MARK = "_trace"

logger = logging.getLogger(__name__)


class StepCollector:
    """ Class for `pytest-bdd` steps dynamic collection. """

    def __init__(self, step_prefixes: Optional[StepPrefixesModel]) -> None:
        self._step_prefixes = step_prefixes
        self._steps: Dict[FeatureTypeName, Dict[str, List[str]]] = {}

    @staticmethod
    def _is_bdd_step(fixture: FixtureDef[Any]) -> bool:
        return (
            isinstance(fixture.argname, str)
            and fixture.argname.startswith(_PYTESTBDD_FIXTURE_MARK)
            and not fixture.argname.endswith(_PYTESTBDD_FIXTURE_TRACE_MARK)
        )

    @classmethod
    def _get_pytestbdd_steps(cls, session: Session) -> Tuple[FixtureDef[Any]]:
        return cast(
            Tuple[FixtureDef[Any]],
            sorted(
                (
                    fx
                    for fx_list in session._fixturemanager._arg2fixturedefs.values()
                    for fx in fx_list
                    if cls._is_bdd_step(fx)
                ),
                key=attrgetter("func.step_type"),
                reverse=True,
            ),
        )

    def _compile_full_step_name(self, fixture_name: str, step_type: str) -> str:
        prefix = step_type.title()
        if self._step_prefixes is not None:
            prefix = self._step_prefixes.dict()[step_type.upper()].strip()
        return f"{prefix} {fixture_name}"

    def _compile_steps_dict(self, steps: Tuple[FixtureDef[Any]]) -> Dict[str, List[str]]:
        return {
            step_type: [
                self._compile_full_step_name(fixture_name=f.func.parser.name, step_type=step_type)  # type: ignore
                for f in steps
                if f.func.step_type == step_type  # type: ignore
            ]
            for step_type in default_types.STEP_TYPES
        }

    def collect_steps(self, session: Session, feature_type: FeatureTypeName) -> None:
        step_fixtures = self._get_pytestbdd_steps(session)
        steps_dict = self._compile_steps_dict(step_fixtures)
        if steps_dict:
            logger.debug("Loaded steps dict:\n%s", steps_dict)
        self._steps[feature_type] = steps_dict

    def get_steps(self, feature_type: FeatureTypeName) -> Dict[str, List[str]]:
        return self._steps.get(feature_type, {})



from typing import Mapping, Type

from overhave.entities.authorization.manager import (
    DefaultAdminAuthorizationManager,
    IAdminAuthorizationManager,
    LDAPAdminAuthorizationManager,
    SimpleAdminAuthorizationManager,
)
from overhave.entities.authorization.strategies import AuthorizationStrategy

AUTH_STRATEGY_TO_MANAGER_MAPPING: Mapping[AuthorizationStrategy, Type[IAdminAuthorizationManager]] = {
    AuthorizationStrategy.SIMPLE: SimpleAdminAuthorizationManager,
    AuthorizationStrategy.DEFAULT: DefaultAdminAuthorizationManager,
    AuthorizationStrategy.LDAP: LDAPAdminAuthorizationManager,
}



from copy import copy
from typing import Dict, Optional, Tuple, cast

from transliterate.base import TranslitLanguagePack


class TranslitPack:
    """ Class for feature file name's transliteration before saving. """

    def __init__(
        self,
        language_code: Optional[str],
        language_name: Optional[str],
        pre_processor_mapping: Optional[Dict[str, str]],
        mapping: Tuple[str, str],
    ) -> None:
        pack = copy(TranslitLanguagePack)
        pack.language_code = language_code
        pack.language_name = language_name
        pack.pre_processor_mapping = pre_processor_mapping
        pack.mapping = mapping
        self._transliterator = pack()

    def translate(self, text: str) -> str:
        return cast(str, self._transliterator.translit(text.strip().lower()).strip("_"))



# flake8: noqa
from .execution_manager import ITestExecutionManager, TestExecutionManager
from .settings import OverhaveProjectSettings, OverhaveTestSettings
from .step_collector import StepCollector
from .test_runner import PytestRunner



import logging
from typing import cast

import werkzeug
from flask_admin import expose
from flask_login import current_user
from wtforms import ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured

logger = logging.getLogger(__name__)


class EmulationRunView(ModelViewConfigured):
    """ View for :class:`EmulationRun` table. """

    details_template = "emulation_run_detail.html"

    can_view_details = True
    can_edit = False
    column_list = (
        "id",
        "emulation.name",
        "initiated_by",
        "status",
        "created_at",
    )
    column_filters = ("emulation.name", "emulation.created_by", "status")
    column_searchable_list = ("emulation.name", "emulation.created_by")

    column_details_list = (
        "emulation.name",
        "id",
        "status",
        "changed_at",
        "port",
        "traceback",
        "created_at",
        "initiated_by",
    )
    column_descriptions = dict(status="Emulation run status", initiated_by="Initiator of emulation start")

    def on_model_delete(self, model: db.EmulationRun) -> None:
        if not (current_user.login == model.emulation.created_by or current_user.role == db.Role.admin):
            raise ValidationError("Only test run initiator could delete test run result!")

    @expose("/details/", methods=("GET", "POST"))
    def details_view(self) -> werkzeug.Response:
        return cast(werkzeug.Response, super().details_view())

    @staticmethod
    def redirect_allowed(model: db.EmulationRun) -> bool:
        return model.port is not None



# flake8: noqa
from .http import (
    OverhaveStashClientSettings,
    StashBranch,
    StashErrorResponse,
    StashHttpClient,
    StashHttpClientConflictError,
    StashPrCreationResponse,
    StashProject,
    StashPrRequest,
    StashRepository,
    StashReviewer,
    StashReviewerInfo,
)
from .redis import (
    AnyRedisTask,
    BaseRedisTask,
    EmulationData,
    EmulationTask,
    PublicationData,
    PublicationTask,
    RedisConsumer,
    RedisConsumerRunner,
    RedisProducer,
    RedisStream,
    TestRunData,
    TestRunTask,
    TRedisTask,
)
from .s3 import OverhaveS3Bucket, OverhaveS3ManagerSettings, S3Manager



import logging
from typing import Any, Callable

import flask
import werkzeug as werkzeug
from flask_admin import expose
from flask_login import current_user
from sqlalchemy.exc import StatementError
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured

logger = logging.getLogger(__name__)


def view_wrapper(function_view: Callable[[Any], werkzeug.Response]) -> Callable[[Any], werkzeug.Response]:
    def wrapper(obj: Any) -> werkzeug.Response:
        try:
            return function_view(obj)
        except StatementError:
            flask.flash("Unsupported symbols in tag name!")
            return flask.redirect(flask.request.url)

    return wrapper


class TagsView(ModelViewConfigured):
    """ View for :class:`Feature` table. """

    can_view_details = False

    column_list = (
        "id",
        "value",
        "created_at",
        "created_by",
    )

    form_excluded_columns = ("created_at",)

    def on_model_change(self, form: Form, model: db.Tags, is_created: bool) -> None:
        if not is_created:
            if current_user.login == model.created_by or current_user.role == db.Role.admin:
                return
            raise ValidationError("Only tag creator or administrator could edit it!")
        model.created_by = current_user.login

    def on_model_delete(self, model: db.Tags) -> None:
        if not (current_user.login == model.created_by or current_user.role == db.Role.admin):
            raise ValidationError("Only author or administrator could delete tags!")

    @expose("/edit/", methods=("GET", "POST"))
    @view_wrapper
    def edit_view(self) -> Any:
        return super().edit_view()

    @expose("/new/", methods=("GET", "POST"))
    @view_wrapper
    def crate_view(self) -> Any:
        return super().create_view()



import logging

import flask
from flask_wtf import FlaskForm as Form
from werkzeug import Response
from wtforms import PasswordField, StringField, validators

from overhave import db
from overhave.entities.authorization.manager import IAdminAuthorizationManager

logger = logging.getLogger(__name__)

_INVALID_AUTH_MSG = "Specified username '{username}' and password pair is invalid!"


class LoginForm(Form):
    """ Form for user authorization. """

    username: StringField = StringField(
        "Username",
        validators=[validators.input_required(message="Field required!")],
        render_kw={"placeholder": "Username", "icon": "glyphicon-user"},
    )
    password: PasswordField = PasswordField(
        "Password", render_kw={"placeholder": "Password", "icon": "glyphicon-certificate"},
    )

    def __init__(self, auth_manager: IAdminAuthorizationManager) -> None:
        super().__init__()
        self._auth_manager = auth_manager

    def get_user(self) -> db.BaseUser:
        authorized_user = self._auth_manager.authorize_user(username_field=self.username, password_field=self.password)
        if authorized_user is None:
            raise validators.ValidationError(_INVALID_AUTH_MSG.format(username=self.username.data))
        return authorized_user

    @staticmethod
    def flash_and_redirect(flash_msg: str) -> Response:
        flask.flash(flash_msg, category="error")
        return flask.redirect(flask.url_for("admin.login"))



import json
import logging
import re
from functools import cached_property
from typing import Any, Dict, List, Optional

import flask
import werkzeug
from flask_admin import expose
from flask_admin.model import InlineFormAdmin
from flask_login import current_user
from markupsafe import Markup
from wtforms import Field, TextAreaField, ValidationError
from wtforms.widgets import HiddenInput

from overhave import db
from overhave.admin.views.base import ModelViewConfigured
from overhave.factory import get_admin_factory
from overhave.transport import TestRunData, TestRunTask

logger = logging.getLogger(__name__)

_SCENARIO_PREFIX = "scenario-0"


class ScenarioTextWidget(HiddenInput):
    """ Widget to override scenario view. """

    def __call__(self, field: Field, **kwargs: Any) -> Any:
        widget_name = type(self).__name__
        logger.debug("%s field id: '%s'", widget_name, field.id)
        inputs = super().__call__(field, **kwargs)
        logger.debug("%s inputs: '%s'", widget_name, inputs)
        return inputs + Markup(f'<div class="editor" ref="{field.id}" id="{field.id}-editor"></div>')


class ScenarioTextAreaField(TextAreaField):
    """ Field to override scenario view. """

    widget = ScenarioTextWidget()


class ScenarioInlineModelForm(InlineFormAdmin):
    """ Form to override scenario view. """

    form_overrides = dict(text=ScenarioTextAreaField)
    form_excluded_columns = ("created_at", "test_runs")


class FeatureView(ModelViewConfigured):
    """ View for :class:`Feature` table. """

    can_view_details = False

    inline_models = (ScenarioInlineModelForm(db.Scenario),)

    create_template = "feature_create.html"
    edit_template = "feature_edit.html"

    column_list = (
        "id",
        "name",
        "feature_type",
        "feature_tags",
        "task",
        "author",
        "created_at",
        "last_edited_by",
        "released",
    )
    form_excluded_columns = (
        "created_at",
        "last_edited_by",
        "released",
        "versions",
        "feature_tags.value",
    )

    column_searchable_list = [
        "id",
        "name",
        "task",
        "author",
        "last_edited_by",
        "feature_tags.value",
    ]
    column_filters = ("name", "feature_type", "last_edited_by", "author", "created_at", "feature_tags.value")
    column_sortable_list = ("id", "name", "author", "last_edited_by")
    column_labels = {"feature_tags.value": "Tags"}

    _task_pattern = re.compile(r"\w+[-]\d+")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def _validate_tasks(cls, tasks: List[str]) -> None:
        for task in tasks:
            if cls._task_pattern.match(task):
                continue
            raise ValidationError(
                f"Incorrect format of task specification: '{task}'! "
                "Supported: <PROJECT>-<NUMBER>, for example 'PRJ-1234'."
            )

    def on_model_change(self, form, model, is_created) -> None:  # type: ignore
        self._validate_tasks(model.task)
        if is_created:
            model.author = current_user.login
        model.last_edited_by = current_user.login
        model.released = False

    def on_model_delete(self, model) -> None:  # type: ignore
        if not (current_user.login == model.author or current_user.role == db.Role.admin):
            raise ValidationError("Only feature author or administrator could delete feature!")

    @cached_property
    def get_bdd_steps(self) -> Dict[str, Dict[str, List[str]]]:
        factory = get_admin_factory()
        return {
            feature_type: factory.step_collector.get_steps(feature_type)
            for feature_type in factory.feature_extractor.feature_types
        }

    @property
    def browse_url(self) -> Optional[str]:
        browse_url_value = get_admin_factory().context.project_settings.browse_url
        if browse_url_value is not None:
            return browse_url_value.human_repr()
        return None

    @staticmethod
    def _run_test(data: Dict[str, Any], rendered: werkzeug.Response) -> werkzeug.Response:
        scenario_id = data.get(f"{_SCENARIO_PREFIX}-id")
        scenario_text = data.get(f"{_SCENARIO_PREFIX}-text")
        if not scenario_id or not scenario_text:
            flask.flash("Scenario information not requested.", category="error")
            return rendered
        factory = get_admin_factory()
        scenario = factory.scenario_storage.get_scenario(int(scenario_id))
        if scenario is None:
            flask.flash("Scenario does not exist, so could not run test.", category="error")
            return rendered
        test_run_id = factory.test_run_storage.create_test_run(scenario_id=scenario.id, executed_by=current_user.login)
        if not factory.redis_producer.add_task(TestRunTask(data=TestRunData(test_run_id=test_run_id))):
            flask.flash("Problems with Redis service! TestRunTask has not been sent.", category="error")
            return rendered
        logger.debug("Redirect to TestRun details view with test_run_id='%s'...", test_run_id)
        return flask.redirect(flask.url_for("testrun.details_view", id=test_run_id))

    @expose("/edit/", methods=("GET", "POST"))
    def edit_view(self) -> werkzeug.Response:
        rendered: werkzeug.Response = super().edit_view()
        if flask.request.method != "POST":
            return rendered

        data = flask.request.form
        logger.debug("Request data:\n%s", json.dumps(data))
        run_scenario_action = data.get("run")
        if not run_scenario_action:
            logger.debug("Show rendered EditView")
            return rendered

        logger.debug("Process feature 'RUN' request")
        tasks = data.get("task").split(",")  # type: ignore
        self._validate_tasks(tasks=tasks)
        return self._run_test(data, rendered)



from typing import cast

import werkzeug
from flask_admin import expose
from flask_login import current_user
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured


class TestRunView(ModelViewConfigured):
    """ View for :class:`TestRun` table. """

    __test__ = False

    list_template = "test_run_list.html"
    details_template = "test_run_detail.html"
    can_create = False
    can_edit = True
    column_searchable_list = (
        "name",
        "executed_by",
    )
    column_list = (
        "id",
        "name",
        "start",
        "end",
        "executed_by",
        "status",
    )
    column_details_list = (
        "id",
        "name",
        "start",
        "end",
        "executed_by",
        "status",
        "report_status",
        "report",
        "traceback",
    )
    column_filters = (
        "name",
        "start",
        "executed_by",
        "status",
    )

    def on_model_change(self, form: Form, model: db.TestRun, is_created: bool) -> None:
        if not is_created and current_user.role != db.Role.admin:
            raise ValidationError("Only administrator could change test run data!")

    def on_model_delete(self, model: db.TestRun) -> None:
        if not (current_user.login == model.executed_by or current_user.role == db.Role.admin):
            raise ValidationError("Only test run initiator could delete test run result!")

    @expose("/details/", methods=("GET", "POST"))
    def details_view(self) -> werkzeug.Response:
        return cast(werkzeug.Response, super().details_view())



import logging

from requests import HTTPError

from overhave.entities import OverhaveFileSettings, PublisherContext
from overhave.publication.git_publisher import BaseGitVersionPublisherError, GitVersionPublisher
from overhave.publication.stash.settings import OverhaveStashPublisherSettings
from overhave.scenario import FileManager
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.test_execution import OverhaveProjectSettings
from overhave.transport import (
    PublicationTask,
    StashBranch,
    StashErrorResponse,
    StashHttpClient,
    StashHttpClientConflictError,
    StashPrCreationResponse,
    StashPrRequest,
)

logger = logging.getLogger(__name__)


class BaseStashVersionPublisherException(BaseGitVersionPublisherError):
    """ Base exception for :class:`StashVersionPublisher`. """


class NullablePullRequestUrlError(BaseStashVersionPublisherException):
    """ Exception for nullable pull-request in selected Draft. """


class StashVersionPublisher(GitVersionPublisher):
    """ Class for feature version's pull requests management relative to Atlassian Stash API. """

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
        stash_publisher_settings: OverhaveStashPublisherSettings,
        client: StashHttpClient,
    ):
        super().__init__(
            file_settings=file_settings,
            project_settings=project_settings,
            feature_storage=feature_storage,
            scenario_storage=scenario_storage,
            test_run_storage=test_run_storage,
            draft_storage=draft_storage,
            file_manager=file_manager,
        )
        self._stash_publisher_settings = stash_publisher_settings
        self._client = client

    def _save_as_duplicate(self, context: PublisherContext) -> None:
        previous_draft = self._draft_storage.get_previous_feature_draft(context.feature.id)
        if previous_draft.pr_url is None:
            raise NullablePullRequestUrlError(
                "Previous draft with id=%s has not got pull-request URL!", previous_draft.id
            )
        if previous_draft.published_at is None:
            raise RuntimeError
        self._draft_storage.save_response(
            draft_id=context.draft.id,
            pr_url=previous_draft.pr_url,
            published_at=previous_draft.published_at,
            opened=True,
        )

    def publish_version(self, task: PublicationTask) -> None:
        draft_id = task.data.draft_id
        context = self._push_version(draft_id)
        if not isinstance(context, PublisherContext):
            return
        pull_request = StashPrRequest(
            title=context.feature.name,
            description=self._compile_publication_description(context),
            open=True,
            fromRef=StashBranch(id=context.target_branch, repository=self._stash_publisher_settings.repository),
            toRef=self._stash_publisher_settings.target_branch,
            reviewers=self._stash_publisher_settings.get_reviewers(feature_type=context.feature.feature_type.name),
        )
        logger.info("Prepared pull-request: %s", pull_request.json(by_alias=True))
        try:
            response = self._client.send_pull_request(pull_request)
            if isinstance(response, StashPrCreationResponse):
                self._draft_storage.save_response(
                    draft_id=draft_id,
                    pr_url=response.get_pr_url(),
                    published_at=response.created_date,
                    opened=response.open,
                )
                return
            if isinstance(response, StashErrorResponse) and response.duplicate:
                self._save_as_duplicate(context)
                return
            logger.error("Gotten error response from Stash: %s", response)
        except StashHttpClientConflictError:
            logger.exception("Gotten conflict. Try to return last pull-request for Draft with id=%s...", draft_id)
            self._save_as_duplicate(context)
        except HTTPError:
            logger.exception("Got HTTP error while trying to sent pull-request!")



from overhave.entities import TranslitPack

RUSSIAN_TRANSLIT_PACK = TranslitPack(
    language_code="ru",
    language_name="russian",
    pre_processor_mapping={
        "ж": "zh",
        "х": "kh",
        "ц": "ts",
        "ч": "ch",
        "ш": "sh",
        "щ": "shch",
        "ю": "yu",
        "я": "ya",
        "ъ": "",
        "ь": "",
        " ": "_",
        ",": "",
        ".": "",
        "(": "",
        ")": "",
        ":": "",
        "?": "",
    },
    mapping=("абвгдеёзийклмнопрстуфыэ₽", "abvgdeeziyklmnoprstufye_"),
)



import logging
from pprint import pformat
from typing import List, Optional, cast

import sqlalchemy.orm as so
from wtforms import PasswordField, StringField

from overhave import db
from overhave.entities.authorization.manager.base import BaseAdminAuthorizationManager
from overhave.entities.authorization.manager.ldap.authenticator import LDAPAuthenticator
from overhave.entities.authorization.settings import OverhaveAuthorizationSettings

logger = logging.getLogger(__name__)


class LDAPEmptyAdminGroupError(ValueError):
    """ Exception for situation with empty LDAP admin group. """

    pass


class LDAPAdminAuthorizationManager(BaseAdminAuthorizationManager):
    """ Class for user authorization via LDAP.

    Manager authorize users using remote LDAP server. Every user should use his LDAP credentials.
    LDAP server returns user groups. If user in default 'admin' group or his groups list contains admin group - user
    will be authorized. If user already placed in database - user will be authorized too. No one password stores.
    """

    def __init__(self, settings: OverhaveAuthorizationSettings, ldap_authenticator: LDAPAuthenticator):
        if settings.admin_group is None:
            raise LDAPEmptyAdminGroupError("Admin group could not be empty!")
        super().__init__(settings)
        self._ldap_authenticator = ldap_authenticator

    def _reassign_role_if_neccessary(self, session: so.Session, user: db.UserRole, user_groups: List[str]) -> None:
        if self._settings.admin_group is not None and self._settings.admin_group in user_groups:
            user.role = db.Role.admin
            session.add(user)
            session.commit()

    def authorize_user(self, username_field: StringField, password_field: PasswordField) -> Optional[db.BaseUser]:
        logger.debug("Try to authorize user '%s'...", username_field.data)
        user_groups = self._ldap_authenticator.get_user_groups(username_field.data, password_field.data)
        if not user_groups:
            logger.debug("LDAP user does not exist!")
            return None
        logger.debug("LDAP user groups: \n %s", pformat(user_groups))
        with db.create_session(expire_on_commit=False) as s:
            db_user = s.query(db.UserRole).filter(db.UserRole.login == username_field.data).one_or_none()
            if db_user is not None:
                self._reassign_role_if_neccessary(session=s, user=db_user, user_groups=user_groups)
                return cast(db.BaseUser, db_user)
            logger.debug("Have not found user with username '%s'!", username_field.data)
            db_groups = s.query(db.GroupRole).filter(db.GroupRole.group.in_(user_groups)).all()
            if db_groups or self._settings.admin_group in user_groups:
                db_user = self._create_user(session=s, username=username_field.data)
                self._reassign_role_if_neccessary(session=s, user=db_user, user_groups=user_groups)
                return cast(db.BaseUser, db_user)
            logger.debug("Received groups are not supplied with db_groups: \n %s", db_groups)
        return None



import abc
import logging
from typing import Optional

import sqlalchemy.orm as so

from overhave import db
from overhave.entities.authorization.manager.abstract import IAdminAuthorizationManager
from overhave.entities.authorization.settings import OverhaveAuthorizationSettings

logger = logging.getLogger(__name__)


class BaseAdminAuthorizationManager(IAdminAuthorizationManager, abc.ABC):
    """ Base class for user authorization. """

    def __init__(self, settings: OverhaveAuthorizationSettings):
        self._settings = settings
        logger.info("Authorization manager: %s", type(self))

    @staticmethod
    def _create_user(
        session: so.Session, username: str, password: Optional[str] = None, role: db.Role = db.Role.user
    ) -> db.UserRole:
        db_user = db.UserRole(login=username, password=password, role=role)
        session.add(db_user)
        session.commit()
        return db_user



import json
import logging
from typing import Optional

import flask
import werkzeug
from flask_admin import expose
from flask_admin.model.helpers import get_mdict_item_or_list
from flask_login import current_user
from wtforms import Form, ValidationError

from overhave import db
from overhave.admin.views.base import ModelViewConfigured
from overhave.factory import get_admin_factory
from overhave.transport import EmulationData, EmulationTask

logger = logging.getLogger(__name__)


class EmulationView(ModelViewConfigured):
    """ View for :class:`Emulation` table. """

    create_template = "emulation_create.html"
    edit_template = "emulation_edit.html"

    can_view_details = False
    column_list = ["id", "name", "test_user.feature_type", "test_user", "created_by"]
    column_searchable_list = ["name", "created_by"]
    form_excluded_columns = ("created_at", "emulation_runs")

    column_labels = {"test_user.feature_type": "Template type"}
    column_descriptions = {
        "name": "Emulation name",
        "test_user.feature_type": "Type of template supported by specified test user",
        "test_user": "Related Test User used for specified emulation",
        "command": "Flags and values for emulation execution",
    }

    @property
    def additional_cmd_description(self) -> Optional[str]:
        return get_admin_factory().context.emulation_settings.emulation_desc_link

    def on_model_change(self, form: Form, model: db.Emulation, is_created: bool) -> None:
        if is_created:
            model.created_by = current_user.login

    def on_model_delete(self, model: db.Emulation) -> None:
        if not (current_user.login == model.created_by or current_user.role == db.Role.admin):
            raise ValidationError("Only emulation item creator or administrator could delete it!")

    @staticmethod
    def _run_emulation(emulation_id: int) -> werkzeug.Response:
        factory = get_admin_factory()
        emulation_run = factory.emulation_storage.create_emulation_run(
            emulation_id=emulation_id, initiated_by=current_user.login
        )
        if not factory.redis_producer.add_task(EmulationTask(data=EmulationData(emulation_run_id=emulation_run.id))):
            flask.flash("Problems with Redis service! EmulationTask has not been sent.", category="error")
            return flask.redirect(flask.url_for("emulation.edit_view", id=emulation_id))
        return flask.redirect(flask.url_for("emulationrun.details_view", id=emulation_run.id))

    @property
    def description_link(self) -> Optional[str]:
        return get_admin_factory().context.emulation_settings.emulation_desc_link

    @expose("/edit/", methods=("GET", "POST"))
    def edit_view(self) -> Optional[werkzeug.Response]:
        data = flask.request.form
        logger.debug("Request data:\n%s", json.dumps(data))

        rendered: werkzeug.Response = super().edit_view()

        emulation_action = data.get("emulate")
        if not emulation_action:
            logger.debug("Show rendered EditView")
            return rendered
        emulation_id = get_mdict_item_or_list(flask.request.args, "id")
        if not emulation_id:
            flask.flash("Please, save emulation template before execution.")
            return rendered

        logger.debug("Seen emulation request")
        return self._run_emulation(emulation_id)



import abc
import logging
import tempfile
from pathlib import Path

from overhave.db import TestRunStatus
from overhave.entities import OverhaveFileSettings, ReportManager, TestExecutorContext
from overhave.scenario import FileManager
from overhave.storage import IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.test_execution.test_runner import PytestRunner
from overhave.transport import TestRunTask

logger = logging.getLogger(__name__)


class ITestExecutionManager(abc.ABC):
    """ Abstract class for test execution management. """

    @abc.abstractmethod
    def execute_test(self, task: TestRunTask) -> None:
        pass


class BaseTestExecutionManagerException(Exception):
    """ Base exception for :class:`TestExecutionManager`. """


class TestRunNotExistsError(BaseTestExecutionManagerException):
    """ Exception for situation with not existing TestRun. """


class ScenarioNotExistsError(BaseTestExecutionManagerException):
    """ Exception for situation with not existing Scenario. """


class TestExecutionManager(ITestExecutionManager):
    """ Class for test execution management. """

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        file_manager: FileManager,
        test_runner: PytestRunner,
        report_manager: ReportManager,
    ):
        self._file_settings = file_settings
        self._feature_storage = feature_storage
        self._scenario_storage = scenario_storage
        self._test_run_storage = test_run_storage
        self._file_manager = file_manager
        self._test_runner = test_runner
        self._report_manager = report_manager

    def _run_test(self, context: TestExecutorContext, alluredir: Path) -> int:
        with self._file_manager.tmp_feature_file(context=context) as feature_file:
            with self._file_manager.tmp_fixture_file(context=context, feature_file=feature_file) as fixture_file:
                return self._test_runner.run(fixture_file=fixture_file.name, alluredir=alluredir.as_posix())

    def _compile_context(self, test_run_id: int) -> TestExecutorContext:
        test_run = self._test_run_storage.get_test_run(test_run_id)
        if test_run is None:
            raise TestRunNotExistsError(f"TestRun with id={test_run_id} does not exist!")
        scenario = self._scenario_storage.get_scenario(test_run.scenario_id)
        if scenario is None:
            raise ScenarioNotExistsError(f"Scenario with id={test_run.scenario_id} does not exist!")
        feature = self._feature_storage.get_feature(scenario.feature_id)
        return TestExecutorContext(feature=feature, scenario=scenario, test_run=test_run,)

    def execute_test(self, task: TestRunTask) -> None:
        test_run_id = task.data.test_run_id
        self._test_run_storage.set_run_status(run_id=test_run_id, status=TestRunStatus.RUNNING)
        ctx = self._compile_context(test_run_id)

        results_dir = Path(tempfile.mkdtemp())
        logger.debug("Allure results directory path: %s", results_dir.as_posix())
        try:
            test_return_code = self._run_test(context=ctx, alluredir=results_dir)
        except Exception as e:
            logger.exception("Error!")
            self._test_run_storage.set_run_status(
                run_id=test_run_id, status=TestRunStatus.INTERNAL_ERROR, traceback=str(e)
            )
            return

        logger.debug("Test returncode: %s", test_return_code)
        if test_return_code == 0:
            self._test_run_storage.set_run_status(run_id=test_run_id, status=TestRunStatus.SUCCESS)
        else:
            self._test_run_storage.set_run_status(
                run_id=test_run_id, status=TestRunStatus.FAILED, traceback="Test run failed!"
            )
        self._report_manager.create_allure_report(test_run_id=test_run_id, results_dir=results_dir)



from flask_admin import Admin
from flask_admin.menu import MenuLink

from overhave.admin.views.index.view import OverhaveIndexView


def get_flask_admin(index_view: OverhaveIndexView) -> Admin:
    admin = Admin(name="Overhave", template_mode="bootstrap3", index_view=index_view,)
    admin.add_link(MenuLink(name="Log out", url="/logout"))
    return admin



from typing import Optional

from overhave.entities import (
    OverhaveDescriptionManagerSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveReportManagerSettings,
    OverhaveScenarioCompilerSettings,
    OverhaveStepContextSettings,
)
from overhave.factory.context.base_context import BaseFactoryContext
from overhave.test_execution import OverhaveProjectSettings, OverhaveTestSettings
from overhave.transport import OverhaveS3ManagerSettings


class OverhaveTestExecutionContext(BaseFactoryContext):
    """ Overhave test run context, based on application BaseSettings.

    This context defines how Overhave test run will work.
    """

    def __init__(
        self,
        compilation_settings: Optional[OverhaveScenarioCompilerSettings] = None,
        description_manager_settings: Optional[OverhaveDescriptionManagerSettings] = None,
        file_settings: Optional[OverhaveFileSettings] = None,
        language_settings: Optional[OverhaveLanguageSettings] = None,
        project_settings: Optional[OverhaveProjectSettings] = None,
        report_manager_settings: Optional[OverhaveReportManagerSettings] = None,
        s3_manager_settings: Optional[OverhaveS3ManagerSettings] = None,
        step_context_settings: Optional[OverhaveStepContextSettings] = None,
        test_settings: Optional[OverhaveTestSettings] = None,
    ) -> None:
        super().__init__(
            compilation_settings=compilation_settings or OverhaveScenarioCompilerSettings(),
            file_settings=file_settings or OverhaveFileSettings(),
            language_settings=language_settings or OverhaveLanguageSettings(),
            project_settings=project_settings or OverhaveProjectSettings(),
            report_manager_settings=report_manager_settings or OverhaveReportManagerSettings(),
            s3_manager_settings=s3_manager_settings or OverhaveS3ManagerSettings(),
            test_settings=test_settings or OverhaveTestSettings(),
        )
        self.description_manager_settings = description_manager_settings or OverhaveDescriptionManagerSettings()
        self.step_context_settings = step_context_settings or OverhaveStepContextSettings()



import abc
import logging
from typing import Optional

import git

from overhave.entities import PublisherContext
from overhave.publication.base_publisher import BaseVersionPublisher, BaseVersionPublisherException

logger = logging.getLogger(__name__)


class BaseGitVersionPublisherError(BaseVersionPublisherException):
    """ Base exception for git version publisher. """


class CurrentBranchEqualToDefaultError(BaseGitVersionPublisherError):
    """ Exception for situation when current branch is equal to default branch. """


class CurrentBranchNotEqualToTargetError(BaseGitVersionPublisherError):
    """ Exception for situation when current branch is not equal to target branch. """


class CommitNotCreatedError(BaseGitVersionPublisherError):
    """ Exception for situation with not created commit. """


class GitVersionPublisher(BaseVersionPublisher, abc.ABC):
    """ Class for feature version's management, based on git. """

    @staticmethod
    def _create_head(name: str, repository: git.Repo) -> git.Head:
        origin: git.Head = repository.remotes.origin
        origin.pull()
        logger.info("Origin successfully pulled")
        if name in repository.heads:
            repository.delete_head(repository.heads[name], force=True)
            logger.warning("Trashed existed branch with name '%s'", name)

        branch = repository.create_head(name)
        logger.info("Created head for branch: %s", branch)
        assert repository.active_branch != branch
        return branch

    def _create_commit_and_push(self, context: PublisherContext, repository: git.Repo, target_branch: git.Head) -> None:
        original_branch = repository.active_branch
        changes_pushed = False
        try:
            target_branch.checkout()
            logger.info("Active branch after checkout: %s", repository.active_branch)

            feature_file = self._file_manager.produce_feature_file(context=context)
            logger.info("Untracked files after feature creation: %s", repository.untracked_files)
            repository.index.add([feature_file.as_posix()])
            logger.info("Untracked files after index update: %s", repository.untracked_files)

            repository.index.commit(
                (
                    f"Added feature {context.feature.name} "
                    f"(id {context.feature.id}) "
                    f"by {context.test_run.executed_by}"
                ),
                skip_hooks=True,
            )
            logger.info("Commit successfully created")
            if repository.active_branch == original_branch:
                raise CurrentBranchEqualToDefaultError(
                    f"Current branch could not be equal to default: '{original_branch}'!"
                )
            if repository.active_branch != target_branch:
                raise CurrentBranchNotEqualToTargetError(
                    f"Current branch should be equal to target: '{target_branch}'!"
                )

            logger.info("Remote origin url: %s", repository.remotes.origin.url)
            repository.git.push("origin", target_branch, force=True)
            logger.info("Changes successfully pushed to remote repository")
            changes_pushed = True
        except (git.GitCommandError, BaseGitVersionPublisherError):
            logger.exception("Error while trying to commit feature file!")

        if original_branch != repository.active_branch:
            logger.info("Current branch is '%s'", repository.active_branch)
            original_branch.checkout()
            logger.info("Branch returned to '%s'", repository.active_branch)

        if not changes_pushed:
            raise CommitNotCreatedError("Commit has not been created!")

    def _push_version(self, draft_id: int) -> Optional[PublisherContext]:
        try:
            repository = git.Repo(self._file_settings.features_dir)
            logger.info("Repository: %s", repository)
            logger.info("Working dir: %s", repository.working_dir)
            logger.info("Active head: %s", repository.active_branch)

            ctx = self._compile_context(draft_id)
            git_branch = self._create_head(repository=repository, name=ctx.target_branch)
            self._create_commit_and_push(context=ctx, repository=repository, target_branch=git_branch)
            return ctx
        except (git.GitCommandError, BaseGitVersionPublisherError):
            logger.exception("Error while trying to push scenario version!")
            return None



# flake8: noqa
from .base import BaseTable, current_session, metadata
from .statuses import EmulationStatus, TestReportStatus, TestRunStatus
from .tables import Draft, Emulation, EmulationRun, Feature, FeatureType, Scenario, Tags, TestRun, TestUser
from .users import BaseUser, GroupRole, Role, UserRole
from .utils import create_session, ensure_feature_types_exist



import click

from overhave.base_settings import DataBaseSettings, LoggingSettings
from overhave.cli.group import overhave
from overhave.factory import ConsumerFactory
from overhave.transport import RedisStream


def _run_consumer(stream: RedisStream) -> None:
    DataBaseSettings().setup_db()
    LoggingSettings().setup_logging()
    ConsumerFactory(stream=stream).runner.run()


@overhave.command(short_help="Run Overhave Redis consumer")
@click.option(
    "-s",
    "--stream",
    type=click.Choice(RedisStream.__members__),
    callback=lambda c, p, v: getattr(RedisStream, v),
    help="Redis stream, which defines application",
)
def consumer(stream: RedisStream) -> None:
    """ Run Overhave Redis consumer. """
    _run_consumer(stream)



import abc

from overhave.transport import PublicationTask


class IVersionPublisher(abc.ABC):
    """ Abstract class for feature version's pull requests management. """

    @abc.abstractmethod
    def publish_version(self, task: PublicationTask) -> None:
        pass



import logging
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

from overhave.entities import IFeatureExtractor, OverhaveFileSettings, OverhaveLanguageSettings, TestExecutorContext
from overhave.scenario.compiler import ScenarioCompiler
from overhave.test_execution.settings import OverhaveProjectSettings

logger = logging.getLogger(__name__)


class FileSavingMixin:
    """ Mixin for files saving. """

    @staticmethod
    def _write_data(file: Any, data: str, entity_name: str) -> None:
        file.write(data)
        file.flush()
        logger.debug("Generated %s file: '%s'", entity_name, file.name)
        assert Path(file.name).exists(), f"{entity_name.upper()} file not created!"

    @staticmethod
    def _get_feature_name(feature_name: str, feature_id: int) -> str:
        return f"{feature_name}_id{feature_id}"


class FileManager(FileSavingMixin):
    """ Class for feature files management. """

    def __init__(
        self,
        project_settings: OverhaveProjectSettings,
        file_settings: OverhaveFileSettings,
        language_settings: OverhaveLanguageSettings,
        feature_extractor: IFeatureExtractor,
        scenario_compiler: ScenarioCompiler,
    ):
        self._language_settings = language_settings
        self._project_settings = project_settings
        self._file_settings = file_settings
        self._feature_extractor = feature_extractor
        self._scenario_compiler = scenario_compiler

    def _compile_feature_file_name_prefix(self, context: TestExecutorContext) -> str:
        feature_name = context.feature.name
        if self._language_settings.translit_pack is not None:
            feature_name = self._language_settings.translit_pack.translate(feature_name)
        return self._get_feature_name(feature_name=feature_name, feature_id=context.feature.id)

    @contextmanager
    def tmp_feature_file(self, context: TestExecutorContext) -> Iterator:  # type: ignore
        run_prefix = self._compile_feature_file_name_prefix(context)
        logger.debug("Feature prefix: '%s'", run_prefix)
        with tempfile.NamedTemporaryFile(
            dir=self._file_settings.tmp_features_dir,
            prefix=f"{run_prefix}_",
            suffix=self._file_settings.feature_suffix,
            mode="w",
        ) as file:
            data = self._scenario_compiler.compile(context=context)
            logger.debug("Scenario file:\n%s", data)
            self._write_data(file=file, data=data, entity_name="feature")
            yield file

    @contextmanager
    def tmp_fixture_file(self, context: TestExecutorContext, feature_file: Any) -> Iterator:  # type: ignore
        with tempfile.NamedTemporaryFile(
            dir=self._file_settings.tmp_fixtures_dir,
            prefix=f"{context.test_run.id}_",
            suffix=self._file_settings.fixture_suffix,
            mode="w",
        ) as file:
            joined_content = "\n".join(self._project_settings.fixture_content)
            data = joined_content.format(feature_file_path=feature_file.name)

            logger.debug("Fixture file:\n%s", data)
            self._write_data(file=file, data=data, entity_name="fixture")
            yield file

    def produce_feature_file(self, context: TestExecutorContext) -> Path:
        feature_file_path: Path = (
            self._feature_extractor.feature_type_to_dir_mapping[context.feature.feature_type.name]
            / f"{self._compile_feature_file_name_prefix(context)}{self._file_settings.feature_suffix}"
        )
        if not feature_file_path.exists():
            logger.info("Create feature file '%s' for commit", feature_file_path.as_posix())
            feature_file_path.touch()
        logger.info("Write scenario to feature file '%s'...", feature_file_path.as_posix())
        with feature_file_path.open("w") as feature_file:
            scenario_text = self._scenario_compiler.compile(context=context)
            logger.debug(scenario_text)
            feature_file.write(scenario_text)
        logger.info("Scenario has successfully written")
        return feature_file_path
